@@ -4,7 +4,7 @@ from cStringIO import StringIO
 from urlparse import urlparse, urlunparse
 from urllib import unquote
 from Cookie import BaseCookie
-from functools import wraps
+from functools import wraps, partial
 from itertools import chain
 from shutil import copyfileobj
 from tempfile import NamedTemporaryFile
@@ -14,8 +14,8 @@ import re
 import webbrowser
 
 from lxml.html import fromstring, tostring
-from lxml.cssselect import CSSSelector
-from lxml.etree import XPath
+from lxml.cssselect import CSSSelector, SelectorSyntaxError
+from lxml.etree import XPath, XPathError
 
 from pesto.testing import MockResponse
 from pesto.request import Request
@@ -48,6 +48,7 @@ class XPathMultiMethod(object):
 
     def __init__(self):
         self.__doc__ = ''
+        self.__name__ = ''
         self.endpoints = []
 
     def __call__(self, *args, **kwargs):
@@ -56,7 +57,8 @@ class XPathMultiMethod(object):
         for xpath, func in self.endpoints:
             if el in xpath(el):
                 return func(*args, **kwargs)
-        raise NotImplementedError("Function not implemented for element %r" % (el,))
+        funcname = self.endpoints[0][1].__name__
+        raise NotImplementedError("Function %s not implemented for element %r" % (self.__name__, el,))
 
     def register(self, xpath, func):
         self.endpoints.append((
@@ -76,6 +78,7 @@ class XPathMultiMethod(object):
             xpath,
             '\n'.join('    %s' % line for line in func.func_doc.split('\n'))
         )
+        self.__name__ = func.__name__
 
 def when(xpath_expr):
     """
@@ -99,7 +102,7 @@ class ElementWrapper(object):
     """
     Wrapper for an ``lxml.etree`` element, providing additional methods useful
     for driving/testing WSGI applications. ``ElementWrapper`` objects are
-    normally created through the ``find``/``findcss`` methods of ``TestAgent``
+    normally created through the ``find``/``css`` methods of ``TestAgent``
     instance::
 
         >>> from pesto.response import Response
@@ -132,11 +135,19 @@ class ElementWrapper(object):
 
     __repr__ = __str__
 
+    def __eq__(self, other):
+        if self.__class__ is not other.__class__:
+            return False
+        return (
+            self.element is other.element
+            and self.agent is other.agent
+        )
+
+
     def __getattr__(self, attr):
         return getattr(self.element, attr)
 
     def __getitem__(self, xpath):
-
         try:
             element = self.element.xpath(xpath)[0]
         except IndexError:
@@ -144,6 +155,16 @@ class ElementWrapper(object):
 
         return self.__class__(self.agent, element)
 
+    def __call__(self, path, flavor='auto', **kwargs):
+        if flavor == 'auto':
+            flavor = guess_expression_flavor(path)
+
+        if flavor == 'css':
+            return self.css(path, **kwargs)
+        else:
+            return self.find(path, **kwargs)
+
+
     def find(self, path, namespaces=None, **kwargs):
         """
         Return elements matching the given xpath expression.
@@ -169,27 +190,27 @@ class ElementWrapper(object):
             return result
 
         return ResultWrapper(
-            ElementWrapper(self.agent, el) for el in result
+            (ElementWrapper(self.agent, el) for el in result),
+            'xpath:' + path
         )
 
-    def findcss(self, selector):
+    def css(self, selector):
         """
         Return elements matching the given CSS Selector (see
         ``lxml.cssselect`` for documentation on the ``CSSSelector`` class.
         """
-        selector = CSSSelector(selector)
+        compiled = CSSSelector(selector)
         return ResultWrapper(
-            ElementWrapper(self.agent, el) for el in selector(self.element)
+            (ElementWrapper(self.agent, el) for el in compiled(self.element)),
+            'css:' + selector
         )
 
-
     def __getitem__(self, path):
         result = self.find(path)
         if len(result) == 0:
             raise ValueError("%r matched no elements" % path)
         return result
 
-
     @when("a[@href]")
     def click(self, follow=True, check_status=True):
         """
@@ -197,31 +218,12 @@ class ElementWrapper(object):
         """
         return self.agent._click(self, follow=follow, check_status=check_status)
 
-    @when("input[@type='checkbox']")
-    def _get_value(self):
+    @when("input[@type='submit' or @type='image']|button[@type='submit' or not(@type)]")
+    def click(self, follow=True, check_status=True):
         """
-        Return the value of the selected checkbox attribute (defaults to ``On``)
+        Alias for submit
         """
-        return self.element.attrib.get('value', 'On')
-
-    @when("input[@type='radio']")
-    def _set_value(self, value):
-        """
-        Set the value of the radio button, by searching for the radio
-        button in the group with the given value and checking it.
-        """
-        found = False
-        for el in self.element.xpath(
-            "./ancestor-or-self::form[1]//input[@type='radio' and @name=$name]",
-            name=self.element.attrib.get('name', '')
-        ):
-            if (el.attrib['value'] == value):
-                el.attrib['checked'] = ""
-                found = True
-            elif 'checked' in el.attrib:
-                del el.attrib['checked']
-        if not found:
-            raise AssertionError("Value %r not present in radio button group %r" % (value, self.element.attrib.get('name')))
+        return self.submit(follow, check_status=check_status)
 
     @when("input[@type='file']")
     def _set_value(self, value):
@@ -261,80 +263,46 @@ class ElementWrapper(object):
         """
         self.element.attrib['value'] = value
 
-    @when("textarea")
-    def _get_value(self):
-        """
-        Return the value of the textarea
-        """
-        return self.element.text
-
-    @when("textarea")
-    def _set_value(self, value):
-        """
-        Set the value of the textarea
-        """
-        self.element.text = value
-
-    @when("select[@multiple]")
-    def _get_value(self):
-        """
-        Return the value of the multiple select
-        """
-        return [item.attrib.get('value') for item in self.element.xpath('./option[@selected]')]
-
-    @when("select[@multiple]")
-    def _set_value(self, values):
-        """
-        Set the value of the multiple select to the given list of values
-        """
-        found = set()
-        values = set(values)
-        for el in self.element.xpath(".//option"):
-            if (el.attrib['value'] in values):
-                el.attrib['selected'] = ""
-                found.add(el.attrib['value'])
-            elif 'selected' in el.attrib:
-                del el.attrib['selected']
-        if found != values:
-            raise AssertionError("Values %r not present in select %r" % (values - found, self.element.attrib.get('name')))
-
-    @when("select")
-    def _get_value(self):
-        """
-        Return the value of the (non-multiple) select box
-        """
-        try:
-            return self.element.xpath('./option[@selected]')[0].attrib['value']
-        except (KeyError, IndexError):
-            return None
-
-    @when("select")
-    def _set_value(self, value):
-        """
-        Set the value of the (non-multiple) select to the given single value
-        """
-        found = False
-        for el in self.element.xpath(".//option"):
-            if (el.attrib.get('value') == value):
-                el.attrib['selected'] = ""
-                found = True
-            elif 'selected' in el.attrib:
-                del el.attrib['selected']
-        if not found:
-            raise AssertionError("Value %r not present in select %r" % (value, self.element.attrib.get('name')))
-
-
-    def __eq__(self, other):
-        if self.__class__ is not other.__class__:
-            return False
-        return (
-            self.element is other.element
-            and self.agent is other.agent
-        )
-
     value = property(_get_value, _set_value)
 
-    @when("input[@type='radio' or @type='checkbox']")
+    @when("textarea|input|select")
+    def input_group(self):
+        """
+        Return the group of inputs sharing the same name attribute
+        """
+        return self.form.find(
+            """.//*[
+                    (local-name() = 'input' or local-name() = 'textarea' or local-name() = 'select')
+                    and (@name='{fieldname}')
+            ]
+            """.format(fieldname=self.attrib['name'])
+        )
+
+    @when("input[@type='checkbox']")
+    def submit_value(self):
+        """
+        Return the value of the selected checkbox element as the user
+        agent would return it to the server in a form submission.
+        """
+        if 'disabled' in self.element.attrib:
+            return None
+        if 'checked' in self.element.attrib:
+            return self.element.attrib.get('value', 'On')
+        return None
+
+    @when("input[@type='radio']")
+    def submit_value(self):
+        """
+        Return the value of the selected radio element as the user
+        agent would return it to the server in a form submission.
+        """
+        if 'disabled' in self.element.attrib:
+            return None
+        if 'checked' in self.element.attrib:
+            return self.element.attrib.get('value', '')
+        return None
+
+    @when("select[@multiple]")
     def submit_value(self):
         """
         Return the value of the selected radio/checkbox element as the user
@@ -342,17 +310,31 @@ class ElementWrapper(object):
         """
         if 'disabled' in self.element.attrib:
             return None
-        if 'checked' in self.element.attrib:
-            return self.value
-        return None
+        return [item.attrib.get('value', item.text) for item in self.element.xpath('./option[@selected]')]
 
-    @when("input[not(@type) or @type != 'submit' and @type != 'image' and @type != 'reset']|select|textarea")
+    @when("select[not(@multiple)]")
+    def submit_value(self):
+        """
+        Return the value of the selected radio/checkbox element as the user
+        agent would return it to the server in a form submission.
+        """
+        if 'disabled' in self.element.attrib:
+            return None
+        try:
+            item = self.element.xpath('./option[@selected]')[0]
+        except IndexError:
+            try:
+                item = self.element.xpath('./option[1]')[0]
+            except IndexError:
+                return None
+        return item.attrib.get('value', item.text)
+
+    @when("input[not(@type) or @type != 'submit' and @type != 'image' and @type != 'reset']")
     def submit_value(self):
         """
         Return the value of any other input element as the user
         agent would return it to the server in a form submission.
         """
-
         if 'disabled' in self.element.attrib:
             return None
         return self.value
@@ -364,6 +346,13 @@ class ElementWrapper(object):
         """
         return None
 
+    @when("textarea")
+    def submit_value(self):
+        """
+        Return the value of any submit/reset input element
+        """
+        return self.element.text
+
     submit_value = property(submit_value)
 
     def _get_checked(self):
@@ -446,23 +435,7 @@ class ElementWrapper(object):
         the selected submit element (input of
         type submit or image, or button with type submit)
         """
-        return self.form.submit(self, follow, check_status=check_status)
-
-    @when("input[@type='submit' or @type='image']|button[@type='submit' or not(@type)]")
-    def click(self, follow=True, check_status=True):
-        """
-        Alias for submit
-        """
-        return self.submit(follow, check_status=check_status)
-
-    @when("input[@type='submit' or @type='image']|button[@type='submit' or not(@type)]")
-    def submit_data(self):
-        """
-        Submit the form, returning a new ``TestAgent`` object, by clicking on
-        the selected submit element (input of
-        type submit or image, or button with type submit)
-        """
-        return self.form.submit_data(self)
+        return self.form.submit(self, follow=follow, check_status=check_status)
 
     @when("form")
     def submit(self, button=None, follow=True, check_status=True):
@@ -482,6 +455,14 @@ class ElementWrapper(object):
             ('POST', 'multipart/form-data'): self.agent.post_multipart,
         }[(method, self.element.attrib.get('enctype'))](path, data, follow=follow, check_status=check_status)
 
+    @when("input[@type='submit' or @type='image']|button[@type='submit' or not(@type)]")
+    def submit_data(self):
+        """
+        Return a list of the data that would be submitted to the server
+        in the format ``[(key, value), ...]``, without actually submitting the form.
+        """
+        return self.form.submit_data(self)
+
     @when("form")
     def submit_data(self, button=None):
         """
@@ -489,6 +470,8 @@ class ElementWrapper(object):
         in the format ``[(key, value), ...]``, without actually submitting the form.
         """
         data = []
+        if isinstance(button, basestring):
+            button = self(button)
 
         if button and 'name' in button.attrib:
             data.append((button.attrib['name'], button.value))
@@ -517,36 +500,170 @@ class ElementWrapper(object):
         return data
 
     @when("form")
-    def fill(self, *pairs, **data):
+    def fill(self, *args, **kwargs):
         """
         Fill the current form with data.
 
-        \*pairs
-            Pairs of (xpath, value)
-
-        \*\*data
-            Dictionary of (fieldname => value) mappings. Values may be lists,
-            in which case multiple fields of the same name will be assigned
-            values.
+        :param \*args: Pairs of ``(selector, value)``
 
+        :param \*\*kwargs: mappings of fieldname to value
+            
+        See the documentation for :meth:`_set_value` implementations
+        for individual form control types to see how values are processed
+        as this varies between text inputs, selects, radio buttons,
+        checkboxes etc
         """
 
-        data = data.items()
-        data = ((key, value if isinstance(value, list) else [value]) for key, value in data)
-        data = (
-            (
-                """
-                    .//*[
-                        (local-name() = 'input' or local-name() = 'textarea' or local-name() = 'select')
-                        and (@name='{fieldname}' or @id='{fieldname}')
-                    ][{index}]
-                """.format(fieldname=fieldname, index=ix+1),
-                item
-            ) for fieldname, values in data for ix, item in enumerate(values)
-        )
+        for selector, value in args:
+            self(selector).fill(value)
 
-        for path, value in chain(data, pairs):
-            self[path].value = value
+        for name, value in kwargs.items():
+            path = """
+            .//*[
+                (local-name() = 'input' or local-name() = 'textarea' or local-name() = 'select')
+                and (@name=$name or @id=$name)
+            ]"""
+            self.find(path, name=name).fill(value)
+
+        return self
+
+    @when("input[@type='checkbox']")
+    def fill(self, values):
+
+        if values is None:
+            values = []
+
+        if isinstance(values, bool):
+            self.checked = True
+
+        elif values and all(isinstance(v, bool) for v in values):
+            # List of bools, eg ``[True, False, True]``
+            for el, checked in zip(self.input_group(), values):
+                if checked:
+                    el.attrib["checked"] = ""
+                elif 'checked' in el.attrib:
+                    del el.attrib['checked']
+
+        else:
+            # List of values, eg ``['1', '23', '8']``
+            found = set()
+            values = set(unicode(v) for v in values)
+            for el in self.input_group():
+                if el.attrib.get('value') in values:
+                    el.attrib['checked'] = ""
+                    found.add(el.attrib['value'])
+                elif 'checked' in el.attrib:
+                    del el.attrib['checked']
+            if found != values:
+                raise AssertionError("Values %r not present in checkbox group %r" % (values - found, self.element.attrib.get('name')))
+
+        return self
+        
+    @when("input[@type='radio']")
+    def fill(self, value):
+        """
+        Set the value of the radio button, by searching for the radio
+        button in the group with the given value and checking it.
+        """
+        if value is not None:
+            value = unicode(value)
+        found = False
+        for el in self.element.xpath(
+            "./ancestor-or-self::form[1]//input[@type='radio' and @name=$name]",
+            name=self.element.attrib.get('name', '')
+        ):
+            if (el.attrib.get('value') == value):
+                el.attrib['checked'] = ""
+                found = True
+            elif 'checked' in el.attrib:
+                del el.attrib['checked']
+        if value is not None and not found:
+            raise AssertionError("Value %r not present in radio button group %r" % (value, self.element.attrib.get('name')))
+        return self
+
+    @when("textarea")
+    def fill(self, value):
+        """
+        Set the value of a textarea control
+        """
+        if value is not None:
+            value = unicode(value)
+        self.element.text = value
+        return self
+
+    @when("input[@type='file']")
+    def fill(self, value):
+        """
+        Set the value of a file input box
+        """
+        if value is None:
+            try:
+                del self.element.attrib['value']
+            except KeyError:
+                pass
+        else:
+            self.value = value
+        return self
+
+    @when("input")
+    def fill(self, value):
+        """
+        Set the value of a (text, password, ...) input box
+        """
+        if value is None:
+            try:
+                del self.element.attrib['value']
+            except KeyError:
+                pass
+        else:
+            self.value = unicode(value)
+        return self
+
+    @when("select[@multiple]")
+    def fill(self, values):
+        """
+        Set the values of a multiple select box
+
+        :param values: list of values to be selected
+        """
+        options = self.element.xpath('.//option')
+        if all(isinstance(v, bool) for v in values):
+            values = [
+                opt.attrib.get('value', opt.text)
+                for selected, opt in zip(values, options)
+            ]
+
+        found = set()
+        values = set(unicode(v) for v in values)
+        for opt in options:
+            value  = opt.attrib.get('value', opt.text)
+            if value in values:
+                opt.attrib['selected'] = ""
+                found.add(value)
+            elif 'selected' in opt.attrib:
+                del opt.attrib['selected']
+        if found != values:
+            raise AssertionError("Values %r not present in select %r" % (values - found, self.element.attrib.get('name')))
+        return self
+
+    @when("select[not(@multiple)]")
+    def fill(self, value):
+        """
+        Set the values of a multiple select box
+
+        :param values: list of values to be selected
+        """
+        if value is not None:
+            value = unicode(value)
+        found = False
+        for opt in self.element.xpath('.//option'):
+            if opt.attrib.get('value', opt.text) == value:
+                opt.attrib['selected'] = ""
+                found = True
+            elif 'selected' in opt.attrib:
+                del opt.attrib['selected']
+        if not found and value is not None:
+            raise AssertionError("Value %r not present in select %r" % (value, self.element.attrib.get('name')))
         return self
 
     def html(self):
@@ -563,8 +680,9 @@ class ElementWrapper(object):
 
     def striptags(self):
         """
-        Strip tags out of ``lxml.html`` document ``node``, just leaving behind
-        text. Normalize all sequences of whitespace to a single space.
+        Strip tags out of the element and its children to leave only the
+        textual content. Normalize all sequences of whitespace to a single
+        space.
 
         Use this for simple text comparisons when testing for document content
 
@@ -583,8 +701,8 @@ class ElementWrapper(object):
             for subnode in node:
                 for text in _striptags(subnode):
                     yield text
-            if node.tail:
-                yield node.tail
+                if subnode.tail:
+                    yield subnode.tail
         return re.sub(r'\s\s*', ' ', ''.join(_striptags(self.element)))
 
     def __contains__(self, what):
@@ -624,8 +742,9 @@ class ResultWrapper(list):
     ``click()``, setting/getting form field values etc).
 
     """
-    def __init__(self, elements):
+    def __init__(self, elements, expr=None):
         super(ResultWrapper, self).__init__(elements)
+        self.__dict__['expr'] = expr
 
     def __getattr__(self, attr):
         return getattr(self[0], attr)
@@ -634,14 +753,41 @@ class ResultWrapper(list):
         return setattr(self[0], attr, value)
 
     def __getitem__(self, item):
-        if isinstance(item, int):
-            return super(ResultWrapper, self).__getitem__(item)
-        else:
-            return self[0][item]
+        try:
+            if isinstance(item, int):
+                return super(ResultWrapper, self).__getitem__(item)
+            else:
+                return self[0][item]
+        except IndexError, e:
+            raise IndexError("list index out of range for %r" % (self,))
 
     def __contains__(self, what):
         return self[0].__contains__(what)
 
+    def __repr__(self):
+        return "<ResultWrapper %r>" % (self.__dict__['expr'],)
+
+    def filter_on_text(self, matcher):
+        """
+        Return a new :class:`ResultWrapper` of the elements in ``elements``
+        where applying the function ``matcher`` to the text contained in
+        the element results in a truth value.
+        """
+        return self.__class__(
+            (e for e in self if matcher(e.striptags())),
+            self.expr + " (filtered by %s)" % (matcher)
+        )
+
+    def filter(self, matcher):
+        """
+        Return a new :class:`ResultWrapper` of the elements in ``elements``
+        where applying the function ``matcher`` to the element results in
+        a truth value. """
+        return self.__class__(
+            (e for e in self if matcher(e)),
+            self.expr + " (filtered by %s)" % (matcher)
+        )
+
 class TestAgent(object):
     """
     A ``TestAgent`` object provides a user agent for the WSGI application under
@@ -1113,25 +1259,109 @@ class TestAgent(object):
         """
         return self.root_element.find(path, namespaces, **kwargs)
 
+    def __call__(self, path, flavor='auto', **kwargs):
+        return self.root_element(path, flavor, **kwargs)
+
     def __getitem__(self, path):
         return self.root_element[path]
 
-    def findcss(self, selector):
+    def css(self, selector):
         """
         Return elements matching the given CSS Selector (see
         ``lxml.cssselect`` for documentation on the ``CSSSelector`` class.
         """
-        return self.root_element.findcss(selector)
+        return self.root_element.css(selector)
+
+    def click(self, linkspec, flavor='auto', ignorecase=True, index=0, follow=True, check_status=True, **kwargs):
+        """
+        Click the link matched by ``linkspec``. See :meth:`findlinks` for a
+        description of the link finding parameters
+
+        :param linkspec: specification of the link to be clicked
+
+        :param flavor: if ``css``, ``linkspec`` must be a CSS selector, which must returning one or more links;
+            if ``xpath``, ``linkspec`` must be an XPath expression returning one or more links;
+            any other value will be passed to :meth:`findlinks`.
+
+        :param ignorecase: (see :meth:`findlinks`)
+
+        :param index: index of the link to click in the case of multiple matching links
+        """
+        if flavor == 'css':
+            links = self.css(linkspec, **kwargs)
+        elif flavor == 'xpath':
+            links = self.find(linkspec, **kwargs)
+        else:
+            links = self.findlinks(linkspec, flavor, ignorecase, **kwargs)
+        return links[index].click(follow=follow, check_status=check_status)
+
+    def findlinks(self, linkspec, flavor='auto', ignorecase=True, **kwargs):
+        """
+        Return a :class:`ResultWrapper` of links matched by ``linkspec``.
+
+        :param linkspec: specification of the link to be clicked
+        :param ignorecase: if ``True``, the link search will be case insensitive 
+        :param flavor: one of ``auto``, ``text``, ``contains``, ``startswith``, ``re``
+
+        The ``flavor`` parameter is interpreted according to the following rules:
+
+            - if ``auto``: detect links based on the following criteria:
+
+                - if ``linkspec`` is a regular expression or otherwise has a
+                  ``search`` method, this will be used to match links.
+
+                - if ``linkspec`` is callable, each link will be tested
+                  against it in turn, and the first link that returns True
+                  will be selected.
+
+                - otherwise ``contains`` matching will be used
+
+            - if ``text``: for links where the text of the link is ``linkspec``
+            - if ``contains``: for links where the link text contains ``linkspec``
+            - if ``startswith``: for links where the link text contains ``linkspec``
+            - if ``re``: for links where the text of the link matches ``linkspec``
+        """
+
+        links = self.find('//a')
+
+        if flavor == 'auto':
+            if callable(linkspec):
+                return links.filter(linkspec)
+            elif hasattr(linkspec, 'search') and callable(linkspec.search):
+                return links.filter_on_text(linkspec.search)
+            else:
+                flavor = 'contains'
+
+        if ignorecase and flavor in ('text', 'contains', 'startswith'):
+            linkspec = linkspec.lower()
+            normcase = lambda x: x.lower()
+        else:
+            normcase = lambda x: x
+        
+        if flavor == 'text':
+            matcher = lambda text, l=linkspec: l == normcase(text)
+
+        elif flavor == 'contains':
+            matcher = lambda text, l=linkspec: l in normcase(text)
+
+        elif flavor == 'startswith':
+            matcher = lambda text, l=linkspec: normcase(text).startswith(l)
+
+        elif flavor == 're':
+            flags = re.I if ignorecase else 0
+            matcher = re.compile(linkspec, flags).search
+
+        else:
+            raise AssertionError("bad flavor: " + flavor)
+
+        return links.filter_on_text(matcher)
 
-    def click(self, path, follow=True, check_status=True, **kwargs):
-        return self.find(path, **kwargs).click(follow=follow, check_status=check_status)
 
     def _click(self, element, follow=True, check_status=True):
-        return self.get(
-            element.attrib['href'],
-            follow=follow,
-            check_status=check_status,
-        )
+        href = element.attrib['href']
+        if '#' in href:
+            href = href.split('#')[0]
+        return self.get(href, follow=follow, check_status=check_status)
 
     def follow(self):
         """
@@ -1172,7 +1402,7 @@ class TestAgent(object):
         Return a new TestAgent with all cookies deleted.
         This gives an easy way to test session expiry.
         """
-        return self.__class__(
+        agent = self.__class__(
             self.app,
             self.request,
             self.response,
@@ -1182,6 +1412,8 @@ class TestAgent(object):
             host=self.host,
             port=self.port,
         )
+        agent.cookies = BaseCookie()
+        return agent
 
     def back(self, count=1):
         return self.history[-abs(count)]
@@ -1319,3 +1551,24 @@ def is_html(response):
 def escapeattrib(s):
     return s.replace('&', '&amp;').replace('"', '&quot;')
 
+def guess_expression_flavor(expr):
+    """
+    Try to guess whether ``expr`` is a CSS selector or XPath expression.
+
+    ``css`` is the default value returned for expressions valid in both syntaxes.
+    """
+    try:
+        XPath(expr)
+    except XPathError:
+        return 'css'
+
+    try:
+        CSSSelector(expr)
+    except (AssertionError, SelectorSyntaxError):
+        return 'xpath'
+
+    if '/' in expr:
+        return 'xpath'
+    if '@' in expr:
+        return 'xpath'
+    return 'css'
